/*
 * Coords.h
 *
 *  Created on: 18 mars 2017
 *      Author: davidroussel
 */

#ifndef COORDS_H_
#define COORDS_H_

#include <cstdlib>	// pour size_t
#include <initializer_list>

#include "CoordsException.h"

// Déclaration préalables des fonctions et opérateurs amis
template<typename T, size_t N> class Coords;
template<typename T, size_t N> ostream & operator <<(ostream & out,
                                                     const Coords<T, N> & c);
template<typename T, size_t N> T dataAccess(const Coords<T, N> & c,
                                            const size_t index)
                                                throw (CoordsException);

/**
 * Template d'une coordonnée de dimension N contenant des éléments de type T
 * @tparam T le type des éléments de coordonnée
 * @tparam N le nombre de coordonnées
 */
template<typename T, size_t N = 3>
class Coords
{
	private:
		/**
		 * Le tableau contenant les coordonnées
		 */
		T data[N];

		/**
		 * Index interne utilisé dans les méthodes variadiques setValues
		 */
		size_t selfIndex;

		/**
		 * Constante utilisée pour la comparaison des valeurs (à epsilon près)
		 */
		const static T epsilon;

		/**
		 * Mise en place de la prochaine valeur,
		 * voir le cours : Patrons de classe - Exemple (5/6), p 190
		 * @param value la valeur à mettre en place
		 * @post si l'index selfIndex est valide, la valeur d'index selfIndex
		 * a été mise en place avec la valeur value
		 * en la convertissant en T avec un static_cast<T>
		 * @note doit être implémenté dans le header pour éviter d'avoir
		 * à proto-instancier toutes ses variantes possibles
		 */
		template<typename U>
		void setValues(const U & value)
		{
			if (selfIndex < N)
			{
				data[selfIndex] = static_cast<T>(value);
			}
			selfIndex++;
		}

		/**
		 * Mise en place de multiples valeurs (au moins 1),
		 * voir le cours : Patrons de classe - Exemple (5/6), p 190
		 * @param first la première valeur à mettre en place à l'index index
		 * @param others les autres valeurs à mettre en place à la suite
		 * @note doit être implémenté dans le header pour éviter d'avoir
		 * à proto-instancier toutes ses variantes possibles
		 */
		template<typename U, typename ...V>
		void setValues(const U & first, const V & ... others)
		{
			setValues(first);
			if (sizeof...(others) > 0) // Inutile (peut être utilisé sans le sizeof...)
			{
				setValues(others...);
			}
		}

		/**
		 * Vérifie si au moins l'une des valeurs de la coordonnée est infinie
		 * @return true si l'une au moins des valeurs de la coordonnée
		 * est infinie.
		 * @see #operator *=(const Coords<T, N> & c) throw (CoordsException)
		 * @see #operator /=(const Coords<T, N> & c) throw (CoordsException)
		 * @see #operator *(const Coords<T, N> & c) const throw (CoordsException)
		 * @see #operator /(const Coords<T, N> & c) const throw (CoordsException)
		 * @see #dotProduct(const Coords<T, N> & c) const throw (CoordsException)
		 * @see #scale(const T & scale) throw (CoordsException)
		 */
		bool hasInfinity() const;

	public:
		// --------------------------------------------------------------------
		// Constructeurs / Destructeur
		// --------------------------------------------------------------------
		/**
		 * Constructeur par défaut
		 * @post toutes les valeurs de la coordonée ont été initialisées à
		 * static_cast<T>(0) <==> T(0).
		 */
		Coords();

		/**
		 * Constructeur valué variadique
		 * voir le cours : Patrons de classe - Exemple (2/6), p 187
		 * @param values
		 * @pre le nombre de coordonnées de (value + others) peut être différent
		 * de N,
		 * 	- si < N, seules les sizeof...(value + others) premières valeurs
		 * 	seront initialisées, les autres seront initialisées à T(0)
		 * 	- si > N, les N valeurs seront initialisée et les valeurs
		 * 	supplémentaires ignorées
		 * @post Les valeurs contenues dans (value + others) affectées aux
		 * valeurs de la coordonnée ont été convertie en T avec un static_cast<T>
		 * @code
		 * 	Coord<double, 3> c3('a', 2, 3.0, 4); // c3 = (97.0, 2.0, 3.0, 4.0)
		 * 	Coord<double, 4> c4(2); // c4 == (2.0, 0.0, 0.0, 0.0)
		 * @endcode
		 * @note doit être implémenté dans le header pour éviter d'avoir
		 * à proto-instancier toutes ses variantes possibles
		 */
		template<typename ...U>
		Coords(const T & value, const U & ... others) // TODO Compléter la liste d'initialisation
		{
			// TODO Compléter ...
		}

		/**
		 * Constructeur à partir d'une initializer_list,
		 * voir cours : Patrons de classe - Exemple (3/6), p 188
		 * @param list la liste des paramètres,
		 * peut avoir une longueur différente de N :
		 * 	- Si > N, seules les N premières valeurs seront utilisées pour
		 * 	affecter les valeurs de la coordonée
		 * 	- Si < N, les valeurs contenues dans la liste d'initialisation sont
		 * 	toutes utilisées et les valeurs restantes de la coordonnée sont
		 * 	initialisées à T(0).
		 * @code
		 * 	Coord<int, 4> c1 = {1, 2, 3, 4}; // c1 == (1, 2, 3, 4)
		 * 	Coord<int, 4> c2 = {2, 3, 4};	// c2 = (2, 3, 4, 0)
		 * @endcode
		 */
		Coords(std::initializer_list<T> list);

		/**
		 * Constructeur de copie
		 * @param c les coordonnées à copier
		 */
		Coords(const Coords<T, N> & c);

		/**
		 * Constructeur de conversion d'un type de coordonnée à un autre
		 * (aussi bien en terme de type T que de nombre N)
		 * @param c la coordonnée contenant M valeurs de type U à copier
		 * @post les M premières coordonnées de c ont été copiées si M <= N,
		 * et le reste initialisé à T(0) et si M > N seules les N premières
		 * coordonnées de c ont été copiées.
		 * @note doit être implémenté dans le header pour éviter d'avoir
		 * à proto-instancier toutes ses variantes possibles
		 * @see #Coords()
		 * @see #operator =(const Coords<U, M> & c)
		 */
		template<typename U, size_t M>
		Coords(const Coords<U, M> & c) // TODO Compléter la liste d'initialisation
		{
			// TODO Compléter ...
		}

		/**
		 * Destructeur
		 */
		virtual ~Coords();
		// --------------------------------------------------------------------
		// Accesseurs / Mutateurs
		// --------------------------------------------------------------------

		/**
		 * Accesseur en lecture à la ième coordonnée
		 * @param i l'index de la coordonnée voulue
		 * @return la ième coordonnée.
		 * @throws CoordException si l'indice est invalide avec comme
		 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
		 */
		T get(const size_t i) const throw (CoordsException);

		/**
		 * Accesseur en lecture/écriture à la ième coordonnée
		 * @param i l'index de la coordonnée voulue
		 * @return une référence vers la ième coordonnée.
		 * @throws CoordException si l'indice est invalide avec comme
		 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
		 */
		T & get(const size_t i) throw (CoordsException);

		/**
		 * Accesseur en lecture du nombre de valeurs dans ce type de coordonnée
		 * @return le nombre de valeurs dans cette coordonnée
		 */
		static size_t size();

		// --------------------------------------------------------------------
		// Opérateurs
		// --------------------------------------------------------------------

		/**
		 * Opérateur d'accès direct en lecture à la ième coordonnée
		 * @param i l'index de la coordonnée voulue
		 * @return la ième coordonnée.
		 * @throws CoordException si l'indice est invalide avec comme
		 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
		 * @see #get(const size_t i) const throw (CoordsException)
		 */
		T operator [](const size_t i) const throw (CoordsException);

		/**
		 * Opérateur d'accès direct en lecture/écriture à la ième coordonnée
		 * @param i l'index de la coordonnée voulue
		 * @return une référence vers la ième coordonnée.
		 * @throws CoordException si l'indice est invalide avec comme
		 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
		 * @see #get(const size_t i) throw (CoordsException);
		 */
		T & operator [](const size_t i) throw (CoordsException);

		/**
		 * Opérateur de copie/conversion
		 * @param c la coordonnée contenant des valeurs de type U à copier
		 * 	- Lorsque M > N seules les N premières valeurs de c sont copiée
		 * 	- Lorsque M < N seules les M premières valeurs sont copiée et les
		 * 	valeurs restantes sont initialisée à static_cast<T>(0).
		 * @return une référence vers soi-même après la copie
		 * @note doit être implémenté dans le header pour éviter d'avoir
		 * à proto-instancier toutes ses variantes possibles
		 */
		template<typename U, size_t M>
		Coords<T, N> & operator =(const Coords<U, M> & c)
		{
			// TODO Compléter ...
			return *this;
		}

		/**
		 * Opérateur de comparaison
		 * @param c la coordonnée à comparer à soi-même
		 * @return true si l'écart entre chaque paire de valeurs de la coordonnée
		 * courante et de la coordonnée c est inférieur à epsilon
		 */
		bool operator ==(const Coords<T, N> & c) const;

		/**
		 * Opérateur de différence
		 * @param c la coordonnée à comparer à soi-même
		 * @return false si la distance entre la Coord courante et c est
		 * inférieure à epsilon
		 * @see #operator ==(const Coords<T, N> & c) const
		 */
		bool operator !=(const Coords<T, N> & c) const;

		/**
		 * Opérateur de self incrémentation
		 * @param c la coordonnée utilisée pour incrémenter les valeurs de la
		 * coordonnée courante
		 * @return une référence vers soi-même
		 */
		Coords<T, N> & operator +=(const Coords<T, N> & c);

		/**
		 * Opérateur de self décrémentation
		 * @param c la coordonnée utilisée pour décrémenter les valeurs de la
		 * coordonnée courante
		 * @return une référence vers soi-même
		 */
		Coords<T, N> & operator -=(const Coords<T, N> & c);

		/**
		 * Opérateur de self multiplication
		 * @param c la coordonnée utilisée pour multiplier les valeurs de la
		 * coordonnée courante
		 * @return une référence vers soi-même
		 * @throws CoordException si le résultat d'une multiplication produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see hasInfinity() const
		 */
		Coords<T, N> & operator *=(const Coords<T, N> & c)
		    throw (CoordsException);

		/**
		 * Opérateur de self division
		 * @param c la coordonnée utilisée pour diviser les valeurs de la
		 * coordonnée courante
		 * @return une référence vers soi-même
		 * @throws CoordException si le résultat d'une division produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see hasInfinity() const
		 */
		Coords<T, N> & operator /=(const Coords<T, N> & c)
		    throw (CoordsException);

		/**
		 * Opérateur d'addition avec une autre coordonnée
		 * @param c l'autre coordonnée à ajouter à la coordonnée courante
		 * @return une nouvelle coordonnée contenant la somme des deux
		 * @see #Coords(const Coords<T, N> & c);
		 * @see #operator +=(const Coords<T, N> & c)
		 */
		Coords<T, N> operator +(const Coords<T, N> & c) const;

		/**
		 * Opérateur de soustraction avec une autre coordonnée
		 * @param c l'autre coordonnée à ajouter à la coordonnée courante
		 * @return une nouvelle coordonnée contenant la somme des deux
		 * @see #Coords(const Coords<T, N> & c);
		 * @see #operator -=(const Coords<T, N> & c)
		 */
		Coords<T, N> operator -(const Coords<T, N> & c) const;

		/**
		 * Opérateur de multiplication avec une autre coordonnée
		 * @param c l'autre coordonnée à ajouter à la coordonnée courante
		 * @return une nouvelle coordonnée contenant la somme des deux
		 * @throws CoordException si le résultat d'une multiplication produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see #Coords(const Coords<T, N> & c);
		 * @see #operator *=(const Coords<T, N> & c)
		 */
		Coords<T, N> operator *(const Coords<T, N> & c) const
		    throw (CoordsException);

		/**
		 * Opérateur de division avec une autre coordonnée
		 * @param c l'autre coordonnée à ajouter à la coordonnée courante
		 * @return une nouvelle coordonnée contenant la somme des deux
		 * @throws CoordException si le résultat d'une division produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see #Coords(const Coords<T, N> & c);
		 * @see #operator /=(const Coords<T, N> & c)
		 */
		Coords<T, N> operator /(const Coords<T, N> & c) const
		    throw (CoordsException);

		/**
		 * Opérateur ami de sortie standard : Affiche (x1, x2, x3, ... xn) sur le
		 * flux de sortie out
		 * @param out le flux de sortie sur lequel écrire
		 * @param c la coordonnée à afficher sur le flux de sortie
		 * @return une référence vers le flux de sortie
		 */
		friend ostream & operator <<<>(ostream & out, const Coords<T, N> & c);

		// --------------------------------------------------------------------
		// Opérations
		// --------------------------------------------------------------------
		/**
		 * Produit scalaire de la coordonnée courante avec la coordonnée c
		 * @param c la coordonnée avec laquelle calculer le produit scalaire
		 * \f[
		 * \overrightarrow{X} \bullet \overrightarrow{Y} = \sum_{i = 0}^{n-1} x_{i} \times y_{i}
		 * \f]
		 * @return la valeur du produit scalaire
		 * @throws CoordException si le résultat d'une division produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see hasInfinity() const
		 */
		T dotProduct(const Coords<T, N> & c) const throw (CoordsException);

		/**
		 * Le carré de la norme euclidienne de cette coordonnée
		 * \f[
		 * \| \overrightarrow{X} \|^{2} = \overrightarrow{X} \bullet \overrightarrow{X}
		 * \f]
		 * @return la norme euclidienne au carré de cette coordonnée
		 * @see #dotProduct(const Coords<T, N> & c) const
		 */
		T norm2() const;

		/**
		 * Translation de la coordonnée courante
		 * @param c la coordonnée à ajouter à la coordonnée courante
		 * @return une référence vers la coordonnée courante translatée.
		 * @see operator +=(const Coords<T, N> & c)
		 */
		Coords<T, N> & translate(const Coords<T, N> & c);

		/**
		 * Homothétie de la coordonnée courante
		 * @param scale le facteur d'échelle à applique à chaque valeur de la
		 * coordonnée
		 * @return une référence vers la coordonnée modifiée
		 * @throws CoordException si le résultat d'une multiplication produit un
		 * nombre infini avec comme cause
		 * CoordsException::ExceptionCause::INFINITY
		 * @see hasInfinity() const
		 */
		Coords<T, N> & scale(const T & scale) throw (CoordsException);

		// --------------------------------------------------------------------
		// Fonctions amies (pour le test)
		// --------------------------------------------------------------------
		/**
		 * Fonction amie pour accéder au contenu d'une Coords<T, N> sans avoir
		 * à utiliser les accesseurs dans les tests
		 * @param c la coordonnée dont on veut une valeur
		 * @param index l'index de la valeur que l'on souhaite
		 * @return la valeur contenue dans la coordonnée à l'index index.
		 * @throws CoordsExcepetion si l'index est invalide avec comme
		 * cause CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS
		 */
		friend T dataAccess<>(const Coords<T, N> & c, const size_t index)
		    throw (CoordsException);
};

#endif /* COORDS_H_ */
