/*
 * TestCoords.cpp
 *
 *  Created on: 18 mars 2017
 *      Author: davidroussel
 */

#include "Coords.h"

#include "Coords.h"
#include "CoordsException.h"

#include <csignal>	// for signal
#include <sstream>	// for std::[i/o]stringstream
#include <string>	// for std::string
#include <utility>	// for std::move & std::pair
#include <random>	// for random numbers dist. & generator
#include <limits>	// for numeric_limits
using namespace std;

#include <boost/test/included/unit_test.hpp>
#include <boost/test/test_case_template.hpp>
#include <boost/test/output_test_stream.hpp>
using boost::test_tools::output_test_stream;
#include <boost/mpl/list.hpp>
using namespace boost::unit_test;

size_t nbCoordsMin = 2;
size_t nbCoordsMax = 4;
size_t nbNumericTests = 100;
double tol = 0.0001; // Tolerance Percentage when checking numeric value

/**
 * Définition d'un pointeur de fonction pour un handler à un argument entier
 */
typedef void (*SignalHandlerPointer)(int);

/**
 * Signal handler déclenché lors d'un
 * 	- SIGSEGV si l'on tente d'accéder à un pointeur normal détruit par
 * 		un pointeur intelligent
 * 	- SIGABRT si l'on tente d'accéder à un objet pointé NULL
 * @param signal le numéro du signal
 */
void signalHandler(int signal)
{
	switch (signal)
	{
		case SIGSEGV:
			throw CoordsException(CoordsException::ExceptionCause::ILLEGAL_MEMORY_ACCESS, "Illegal memory access");
			break;
		case SIGABRT:
			throw CoordsException(CoordsException::ExceptionCause::ABORT, "Abort requested");
			break;
		default:
			break;
	}
}

/**
 * Creates a random number generator with values contained within
 * [minVal ... maxVal]
 * @tparam T the type of values to generate
 * @param minVal the minimum value
 * @param maxVal the maximum value
 * @return a function returning the next random number
 */
template<typename T>
function<T(void)> getRandomGenerator(const T & minVal = static_cast<T>(-2147483647),
                                     const T & maxVal = static_cast<T>(2147483648))
{
	// clog << "Generating values between" << minVal << " and " << maxVal << endl;
	default_random_engine engine;
	function<T(void)> random;
	if (typeid(T) == typeid(size_t))
	{
		random = bind(uniform_int_distribution<size_t>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(int))
	{
		random = bind(uniform_int_distribution<int>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(float))
	{
		random = bind(uniform_real_distribution<float>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(double))
	{
		random = bind(uniform_real_distribution<double>(minVal, maxVal), engine);
	}
	else
	{
		cerr << "getRandomGenerator() : Unknown value type" << typeid(T).name()
		     << endl;
	}

	return random;
}

/**
 * Test de la méthode size
 * @see size_t Coords<T, N>::size() const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Size, T)
{
	Coords<T, 2> c2;
	BOOST_CHECK_EQUAL(c2.size(), 2);

	Coords<T, 3> c3;
	BOOST_CHECK_EQUAL(c2.size(), 3);

	Coords<T, 4> c4;
	BOOST_CHECK_EQUAL(c2.size(), 4);
}

/**
 * Test du constructeur par défaut de Coords<T, N>
 * @see Coords<T, N>::Coords()
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Coords_, T)
{
	Coords<T, 2> c2;

	for (size_t i = 0; i < c2.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c2, i), static_cast<T>(0.0));
	}

	Coords<T, 3> c3;

	for (size_t i = 0; i < c3.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c3, i), static_cast<T>(0.0));
	}

	Coords<T, 4> c4;

	for (size_t i = 0; i < c4.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c4, i), static_cast<T>(0.0));
	}
}

/**
 * Test du constructeur valué variadique
 * @see Coords<T, N>::Coords(const T & first, Const U & ...others)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Coords_Variadic, T )
{
	function <T()> random = getRandomGenerator<T>();
	T values[] = {random(), random(), random(), random()};

	Coords<T, 2> c21(values[0]);
	Coords<T, 2> c22(values[0], values[1]);
	Coords<T, 2> c23(values[0], values[1], values[2]);
	Coords<T, 2> c24(values[0], values[1], values[2], values[3]);
	Coords<T, 2> * tc2[] = { &c21, &c22, &c23, &c24};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc2[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc2[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc2[c]), i), values[i]);
			}
		}
	}

	Coords<T, 3> c31(values[0]);
	Coords<T, 3> c32(values[0], values[1]);
	Coords<T, 3> c33(values[0], values[1], values[2]);
	Coords<T, 3> c34(values[0], values[1], values[2], values[3]);
	Coords<T, 3> * tc3[] = { &c31, &c32, &c33, &c34};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc3[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc3[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc3[c]), i), values[i]);
			}
		}
	}

	Coords<T, 4> c41(values[0]);
	Coords<T, 4> c42(values[0], values[1]);
	Coords<T, 4> c43(values[0], values[1], values[2]);
	Coords<T, 4> c44(values[0], values[1], values[2], values[3]);
	Coords<T, 4> * tc4[] = { &c41, &c42, &c43, &c44};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc4[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc4[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc4[c]), i), values[i]);
			}
		}
	}
}

/**
 * Test du constructeur à partir d'une liste d'initialisation
 * @see Coords<T, N>::Coords(std::initializer_list<T> list);
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Coords_Initializer_List, T )
{
	function<T()> random = getRandomGenerator<T>();
	T values[] = {random(), random(), random(), random()};

	Coords<T, 2> c21 = {values[0]};
	Coords<T, 2> c22 = {values[0], values[1]};
	Coords<T, 2> c23 = {values[0], values[1], values[2]};
	Coords<T, 2> c24 = {values[0], values[1], values[2], values[3]};
	Coords<T, 2> * tc2[] = { &c21, &c22, &c23, &c24};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc2[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc2[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc2[c]), i), values[i]);
			}
		}
	}

	Coords<T, 3> c31 = {values[0]};
	Coords<T, 3> c32 = {values[0], values[1]};
	Coords<T, 3> c33 = {values[0], values[1], values[2]};
	Coords<T, 3> c34 = {values[0], values[1], values[2], values[3]};
	Coords<T, 3> * tc3[] = { &c31, &c32, &c33, &c34};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc3[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc3[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc3[c]), i), values[i]);
			}
		}
	}

	Coords<T, 4> c41 = {values[0]};
	Coords<T, 4> c42 = {values[0], values[1]};
	Coords<T, 4> c43 = {values[0], values[1], values[2]};
	Coords<T, 4> c44 = {values[0], values[1], values[2], values[3]};
	Coords<T, 4> * tc4[] = { &c41, &c42, &c43, &c44};

	for (size_t c = 0; c < 4; c++)
	{
		for (size_t i = 0; i < tc4[c]->size(); i++)
		{
			if (i > c)
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc4[c]), i), static_cast<T>(0));
			}
			else
			{
				BOOST_CHECK_EQUAL(dataAccess(*(tc4[c]), i), values[i]);
			}
		}
	}
}

/**
 * test du constructeur de copie
 * @see Coords<T, N>::Coords(const Coords<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( CoordsOfCoords, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	T values[] = {random(), random(), random(), random()};

	Coords<T, 2> c21(values[0], values[1], values[2], values[3]);
	Coords<T, 2> c22(c21);
	for (size_t i = 0; i < c22.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c22, i), dataAccess(c21, i));
	}

	Coords<T, 2> c23 = c21;
	for (size_t i = 0; i < c23.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c23, i), dataAccess(c22, i));
	}

	Coords<T, 3> c31(values[0], values[1], values[2], values[3]);
	Coords<T, 3> c32(c31);
	for (size_t i = 0; i < c32.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c32, i), dataAccess(c31, i));
	}

	Coords<T, 3> c33 = c31;
	for (size_t i = 0; i < c33.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c33, i), dataAccess(c32, i));
	}

	Coords<T, 4> c41(values[0], values[1], values[2], values[3]);
	Coords<T, 4> c42(c41);
	for (size_t i = 0; i < c42.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c42, i), dataAccess(c41, i));
	}

	Coords<T, 4> c43 = c41;
	for (size_t i = 0; i < c43.size(); i++)
	{
		BOOST_CHECK_EQUAL(dataAccess(c43, i), dataAccess(c42, i));
	}
}

/**
 * test du constructeur de trans-copie
 * @see template <U, M> Coords<T, N>::Coords(const Coords<U, M> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( CoordsOfCoordsUM, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	for (size_t t = 0; t < nbNumericTests; t++)
	{
		T values[] = {random(), random(), random(), random()};

		Coords<T, 2> c2(values[0], values[1]);
		Coords<T, 3> c3(values[0], values[1], values[2]);
		Coords<T, 4> c4(values[0], values[1], values[2], values[3]);

		Coords<double, 2> cd22(c2);
		Coords<double, 3> cd32(c2);
		Coords<double, 4> cd42(c2);
		Coords<float, 2> cf22(c2);
		Coords<float, 3> cf32(c2);
		Coords<float, 4> cf42(c2);
		Coords<int, 2> ci22(c2);
		Coords<int, 3> ci32(c2);
		Coords<int, 4> ci42(c2);

		Coords<double, 2> cd23(c3);
		Coords<double, 3> cd33(c3);
		Coords<double, 4> cd43(c3);
		Coords<float, 2> cf23(c3);
		Coords<float, 3> cf33(c3);
		Coords<float, 4> cf43(c3);
		Coords<int, 2> ci23(c3);
		Coords<int, 3> ci33(c3);
		Coords<int, 4> ci43(c3);

		Coords<double, 2> cd24(c4);
		Coords<double, 3> cd34(c4);
		Coords<double, 4> cd44(c4);
		Coords<float, 2> cf24(c4);
		Coords<float, 3> cf34(c4);
		Coords<float, 4> cf44(c4);
		Coords<int, 2> ci24(c4);
		Coords<int, 3> ci34(c4);
		Coords<int, 4> ci44(c4);

		/*
		 * Check values from c2, c3, c4
		 */
		for (size_t i = 0; i < c2.size(); i++)
		{
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c2, i)), dataAccess(cd22, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c2, i)), dataAccess(cd32, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c2, i)), dataAccess(cd42, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c2, i)), dataAccess(cf22, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c2, i)), dataAccess(cf32, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c2, i)), dataAccess(cf42, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c2, i)), dataAccess(ci22, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c2, i)), dataAccess(ci32, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c2, i)), dataAccess(ci42, i));

			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c3, i)), dataAccess(cd23, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c3, i)), dataAccess(cd33, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c3, i)), dataAccess(cd43, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c3, i)), dataAccess(cf23, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c3, i)), dataAccess(cf33, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c3, i)), dataAccess(cf43, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c3, i)), dataAccess(ci23, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c3, i)), dataAccess(ci33, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c3, i)), dataAccess(ci43, i));

			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd24, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd34, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd44, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf24, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf34, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf44, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci24, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci34, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci44, i));
		}

		for (size_t i = c2.size(); i < c3.size(); i++)
		{
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c3, i)), dataAccess(cd33, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c3, i)), dataAccess(cd43, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c3, i)), dataAccess(cf33, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c3, i)), dataAccess(cf43, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c3, i)), dataAccess(ci33, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c3, i)), dataAccess(ci43, i));

			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd34, i));
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd44, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf34, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf44, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci34, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci44, i));

			BOOST_CHECK_EQUAL(dataAccess(cd32, i), 0.0);
			BOOST_CHECK_EQUAL(dataAccess(cd42, i), 0.0);
			BOOST_CHECK_EQUAL(dataAccess(cf32, i), 0.0f);
			BOOST_CHECK_EQUAL(dataAccess(cf42, i), 0.0f);
			BOOST_CHECK_EQUAL(dataAccess(ci32, i), 0);
			BOOST_CHECK_EQUAL(dataAccess(ci42, i), 0);

		}

		for (size_t i = c3.size(); i < c4.size(); i++)
		{
			BOOST_CHECK_EQUAL(static_cast<double>(dataAccess(c4, i)), dataAccess(cd44, i));
			BOOST_CHECK_EQUAL(static_cast<float>(dataAccess(c4, i)), dataAccess(cf44, i));
			BOOST_CHECK_EQUAL(static_cast<int>(dataAccess(c4, i)), dataAccess(ci44, i));

			BOOST_CHECK_EQUAL(dataAccess(cd42, i), 0.0);
			BOOST_CHECK_EQUAL(dataAccess(cd43, i), 0.0);
			BOOST_CHECK_EQUAL(dataAccess(cf42, i), 0.0f);
			BOOST_CHECK_EQUAL(dataAccess(cf43, i), 0.0f);
			BOOST_CHECK_EQUAL(dataAccess(ci42, i), 0);
			BOOST_CHECK_EQUAL(dataAccess(ci43, i), 0);
		}
	}
}

/**
 * Test de l'opérateur d'accès aux données en lecture seule
 * T Coords<T>::get(size_T &) const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( GetConst, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		T values[] = {random(), random(), random(), random()};

		Coords<T, 2> c21(values[0], values[1]);
		Coords<T, 3> c31(values[0], values[1], values[2]);
		Coords<T, 4> c41(values[0], values[1], values[2], values[3]);

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_EQUAL(c21.get(i), values[i]);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_EQUAL(c31.get(i), values[i]);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_EQUAL(c41.get(i), values[i]);
		}

		/*
		 * We should get an CoordsException if we try to get invalid index
		 */
		size_t i = indexRandom();
		if (i >= c21.size())
		{
			try
			{
				T value = c21.get(i);

				BOOST_FAIL("Getting Value for index " << i << " in " << c21  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c31.size())
		{
			try
			{
				T value = c31.get(i);

				BOOST_FAIL("Getting Value for index " << i << " in " << c31  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c41.size())
		{
			try
			{
				T value = c41.get(i);

				BOOST_FAIL("Getting Value for index " << i << " in " << c41  << " = " << value << "succeded");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
	}
}

/**
 * Test de l'opérateur d'accès aux données en lecture/écriture
 * T & Coords<T>::get(size_T &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Get, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21;
		Coords<T, 3> c31;
		Coords<T, 4> c41;

		for (size_t i = 0; i < c21.size(); i++)
		{
			T value = random();
			c21.get(i) = value;
			BOOST_CHECK_EQUAL(dataAccess(c21, i), value);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			T value = random();
			c31.get(i) = value;
			BOOST_CHECK_EQUAL(dataAccess(c31, i), value);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			T value = random();
			c41.get(i) = value;
			BOOST_CHECK_EQUAL(dataAccess(c41, i), value);
		}

		/*
		 * We should get an CoordsException if we try to get invalid index
		 */
		size_t i = indexRandom();
		if (i >= c21.size())
		{
			try
			{
				T value = random();
				c21.get(i) = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c21  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c31.size())
		{
			try
			{
				T value = random();
				c31.get(i) = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c31  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c41.size())
		{
			try
			{
				T value = random();
				c41.get(i) = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c41  << " = " << value << "succeded");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
	}
}

/**
 * Test de l'opérateur d'accès aux données en lecture seule
 * T Coords<T>::operator [](size_T &) const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorBracketConst, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		T values[] = {random(), random(), random(), random()};

		Coords<T, 2> c21(values[0], values[1]);
		Coords<T, 3> c31(values[0], values[1], values[2]);
		Coords<T, 4> c41(values[0], values[1], values[2], values[3]);

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_EQUAL(c21[i], values[i]);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_EQUAL(c31[i], values[i]);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_EQUAL(c41[i], values[i]);
		}

		/*
		 * We should get an CoordsException if we try to get invalid index
		 */
		size_t i = indexRandom();
		if (i >= c21.size())
		{
			try
			{
				T value = c21[i];

				BOOST_FAIL("Getting Value for index " << i << " in " << c21  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c31.size())
		{
			try
			{
				T value = c31[i];

				BOOST_FAIL("Getting Value for index " << i << " in " << c31  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c41.size())
		{
			try
			{
				T value = c41[i];

				BOOST_FAIL("Getting Value for index " << i << " in " << c41  << " = " << value << "succeded");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
	}
}

/**
 * Test de l'opérateur d'accès aux données en lecture/écriture
 * T & Coords<T>::operator [](size_T &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorBracket, T )
{
	function<T(void)> random = getRandomGenerator<T>();
	function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21;
		Coords<T, 3> c31;
		Coords<T, 4> c41;

		for (size_t i = 0; i < c21.size(); i++)
		{
			T value = random();
			c21[i] = value;
			BOOST_CHECK_EQUAL(dataAccess(c21, i), value);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			T value = random();
			c31[i] = value;
			BOOST_CHECK_EQUAL(dataAccess(c31, i), value);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			T value = random();
			c41[i] = value;
			BOOST_CHECK_EQUAL(dataAccess(c41, i), value);
		}

		/*
		 * We should get an CoordsException if we try to get invalid index
		 */
		size_t i = indexRandom();
		if (i >= c21.size())
		{
			try
			{
				T value = random();
				c21[i] = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c21  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c31.size())
		{
			try
			{
				T value = random();
				c31[i] = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c31  << " = " << value << "succeded");

			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
		if (i >= c41.size())
		{
			try
			{
				T value = random();
				c41[i] = value;

				BOOST_FAIL("Getting Value for index " << i << " in " << c41  << " = " << value << "succeded");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
			}
		}
	}
}

/**
 * Test de l'opérateur de copie
 * Coords<T>::operator =(const Coords<T> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorCopyOfCoords, T )
{
	function<T(void)> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22;
		Coords<T, 3> c32;
		Coords<T, 4> c42;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_NE(dataAccess(c21, i), dataAccess(c22, i));
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_NE(dataAccess(c31, i), dataAccess(c32, i));
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_NE(dataAccess(c41, i), dataAccess(c42, i));
		}

		c22 = c21;
		c32 = c31;
		c42 = c41;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_EQUAL(dataAccess(c21, i), dataAccess(c22, i));
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_EQUAL(dataAccess(c31, i), dataAccess(c32, i));
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_EQUAL(dataAccess(c41, i), dataAccess(c42, i));
		}
	}
}

/**
 * Test de l'opérateur de comparaison
 * bool Coords<T, N>::operator ==(const Coord<T, N> &) const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorCompare, T )
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22;
		Coords<T, 3> c32;
		Coords<T, 4> c42;

		BOOST_CHECK(!(c21 == c22));

		BOOST_CHECK(!(c31 == c32));

		BOOST_CHECK(!(c41 == c42));

		c22 = c21;
		c32 = c31;
		c42 = c41;

		BOOST_CHECK(c21 == c22);

		BOOST_CHECK(c31 == c32);

		BOOST_CHECK(c41 == c42);
	}
}

/**
 * Test de l'opérateur de différence
 * Coords<T>::operator !=() const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorDifference, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22;
		Coords<T, 3> c32;
		Coords<T, 4> c42;

		BOOST_CHECK(c21 != c22);

		BOOST_CHECK(c31 != c32);

		BOOST_CHECK(c41 != c42);

		c22 = c21;
		c32 = c31;
		c42 = c41;

		BOOST_CHECK(!(c21 != c22));

		BOOST_CHECK(!(c31 != c32));

		BOOST_CHECK(!(c41 != c42));
	}
}

/**
 * Test de l'opérateur de self addition
 * Coords<T, N> & Coords<T, N>::operator +=(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorSelfAddition, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c21Saved(c21);
		Coords<T, 3> c31Saved(c31);
		Coords<T, 4> c41Saved(c41);

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		c21 += c22;
		c31 += c32;
		c41 += c42;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
			                  static_cast<double>(dataAccess(c21Saved, i) + dataAccess(c22, i)),
			                  tol);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
			                  static_cast<double>(dataAccess(c31Saved, i) + dataAccess(c32, i)),
			                  tol);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
			                  static_cast<double>(dataAccess(c41Saved, i) + dataAccess(c42, i)),
			                  tol);
		}
	}
}

/**
 * Test de l'opérateur de self soustraction
 * Coords<T, N> & Coords<T, N>::operator -=(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorSelfSubstraction, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c21Saved(c21);
		Coords<T, 3> c31Saved(c31);
		Coords<T, 4> c41Saved(c41);

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		c21 -= c22;
		c31 -= c32;
		c41 -= c42;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
			                  static_cast<double>(dataAccess(c21Saved, i) - dataAccess(c22, i)),
			                  tol);
		}

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
			                  static_cast<double>(dataAccess(c31Saved, i) - dataAccess(c32, i)),
			                  tol);
		}

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
			                  static_cast<double>(dataAccess(c41Saved, i) - dataAccess(c42, i)),
			                  tol);
		}
	}
}

/**
 * Test de l'opérateur de self multiplication
 * Coords<T, N> & Coords<T, N>::operator *=(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorSelfMultiplication, T)
{
	function<T(void)> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c21Saved(c21);
		Coords<T, 3> c31Saved(c31);
		Coords<T, 4> c41Saved(c41);

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		try
		{
			c21 *= c22;

			for (size_t i = 0; i < c21.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
				                  static_cast<double>(dataAccess(c21Saved, i) * dataAccess(c22, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c21 << " has infinite component in *=" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			c31 *= c32;

			for (size_t i = 0; i < c31.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) * dataAccess(c32, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c31 << " has infinite component in *=" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c41 *= c42;

			for (size_t i = 0; i < c41.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) * dataAccess(c42, i)),
				                  tol);
			}

		}
		catch (CoordsException & e)
		{
			// cerr << c41 << " has infinite component in *=" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
	}

	/*
	 * Checking infinite values
	 */

	T infiniteValue = numeric_limits<T>::max();
	Coords<T, 2> c21(infiniteValue, infiniteValue);
	Coords<T, 3> c31(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	Coords<T, 2> c21Saved(c21);
	Coords<T, 3> c31Saved(c31);
	Coords<T, 4> c41Saved(c41);

	Coords<T, 2> c22(infiniteValue, infiniteValue);
	Coords<T, 3> c32(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c42(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	try
	{
		c21 *= c22;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
			                  static_cast<double>(dataAccess(c21Saved, i) * dataAccess(c22, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << c21 << " has infinite component in *=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
	try
	{
		c31 *= c32;

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
			                  static_cast<double>(dataAccess(c31Saved, i) * dataAccess(c32, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << c31 << " has infinite component in *=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}

	try
	{
		c41 *= c42;

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
			                  static_cast<double>(dataAccess(c41Saved, i) * dataAccess(c42, i)),
			                  tol);
		}

	}
	catch (CoordsException & e)
	{
		// cerr << c41 << " has infinite component in *=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
}

/**
 * Test de l'opérateur de self division
 * Coords<T, N> & Coords<T, N>::operator /=(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorSelfDivision, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c21Saved(c21);
		Coords<T, 3> c31Saved(c31);
		Coords<T, 4> c41Saved(c41);

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		try
		{
			c21 /= c22;
			for (size_t i = 0; i < c21.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
				                  static_cast<double>(dataAccess(c21Saved, i) / dataAccess(c22, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// clog << c21 << " has infinite component in /=" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			c31 /= c32;

			for (size_t i = 0; i < c31.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) / dataAccess(c32, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// clog << c31 << " has infinite component in /=" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c41 /= c42;
			for (size_t i = 0; i < c41.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) / dataAccess(c42, i)),
				                  tol);
			}

		}
		catch (CoordsException & e)
		{
			// clog << c41 << " has infinite component in /=" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
	}

	/*
	 * Checking infinite values
	 */

	T infiniteValue = numeric_limits<T>::max();
	T nullValue = numeric_limits<T>::min();
	Coords<T, 2> c21(infiniteValue, infiniteValue);
	Coords<T, 3> c31(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	Coords<T, 2> c21Saved(c21);
	Coords<T, 3> c31Saved(c31);
	Coords<T, 4> c41Saved(c41);

	Coords<T, 2> c22(nullValue, nullValue);
	Coords<T, 3> c32(nullValue, nullValue, nullValue);
	Coords<T, 4> c42(nullValue, nullValue, nullValue, nullValue);

	try
	{
		c21 /= c22;

		for (size_t i = 0; i < c21.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c21, i)),
			                  static_cast<double>(dataAccess(c21Saved, i) / dataAccess(c22, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << c21 << " has infinite component in /=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
	try
	{
		c31 /= c32;

		for (size_t i = 0; i < c31.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c31, i)),
			                  static_cast<double>(dataAccess(c31Saved, i) / dataAccess(c32, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << c31 << " has infinite component in /=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}

	try
	{
		c41 /= c42;

		for (size_t i = 0; i < c41.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c41, i)),
			                  static_cast<double>(dataAccess(c41Saved, i) / dataAccess(c42, i)),
			                  tol);
		}

	}
	catch (CoordsException & e)
	{
		// cerr << c41 << " has infinite component in /=" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
}

/**
 * Test de l'opérateur d'addition
 * Coords<T, N> Coords<T, N>::operator +(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorAddition, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		Coords<T, 2> c23(c21 + c22);
		Coords<T, 2> c33(c31 + c32);
		Coords<T, 2> c43(c41 + c42);

		for (size_t i = 0; i < c23.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
			                  static_cast<double>(dataAccess(c21, i) + dataAccess(c22, i)),
			                  tol);
		}

		for (size_t i = 0; i < c33.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
			                  static_cast<double>(dataAccess(c31, i) + dataAccess(c32, i)),
			                  tol);
		}

		for (size_t i = 0; i < c43.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
			                  static_cast<double>(dataAccess(c41, i) + dataAccess(c42, i)),
			                  tol);
		}
	}
}

/**
 * Test de l'opérateur de soustraction
 * Coords<T, N> Coords<T, N>::operator -(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorSubstraction, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		Coords<T, 2> c23(c21 - c22);
		Coords<T, 2> c33(c31 - c32);
		Coords<T, 2> c43(c41 - c42);

		for (size_t i = 0; i < c23.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
			                  static_cast<double>(dataAccess(c21, i) - dataAccess(c22, i)),
			                  tol);
		}

		for (size_t i = 0; i < c33.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
			                  static_cast<double>(dataAccess(c31, i) - dataAccess(c32, i)),
			                  tol);
		}

		for (size_t i = 0; i < c43.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
			                  static_cast<double>(dataAccess(c41, i) - dataAccess(c42, i)),
			                  tol);
		}
	}
}

/**
 * Test de l'opérateur de multiplication
 * Coords<T, N> Coords<T, N>::operator *(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorMultiplication, T)
{
	function<T(void)> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		try
		{
			Coords<T, 2> c23(c21 * c22);
			for (size_t i = 0; i < c21.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) * dataAccess(c22, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			Coords<T, 3> c33(c31 * c32);

			for (size_t i = 0; i < c31.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) * dataAccess(c32, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			Coords<T, 4> c43(c41 * c42);

			for (size_t i = 0; i < c41.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) * dataAccess(c42, i)),
				                  tol);
			}

		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
	}

	/*
	 * Checking infinite values
	 */

	T infiniteValue = numeric_limits<T>::max();
	Coords<T, 2> c21(infiniteValue, infiniteValue);
	Coords<T, 3> c31(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	Coords<T, 2> c22(infiniteValue, infiniteValue);
	Coords<T, 3> c32(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c42(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	try
	{
		Coords<T, 2> c23(c21 *c22);

		for (size_t i = 0; i < c23.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
			                  static_cast<double>(dataAccess(c21, i) * dataAccess(c22, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in *" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
	try
	{
		Coords<T, 3> c33(c31 * c32);

		for (size_t i = 0; i < c33.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
			                  static_cast<double>(dataAccess(c31, i) * dataAccess(c32, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in *" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}

	try
	{
		Coords<T, 4> c43(c41 *c42);

		for (size_t i = 0; i < c43.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
			                  static_cast<double>(dataAccess(c41, i) * dataAccess(c42, i)),
			                  tol);
		}

	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in *" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
}

/**
 * Test de l'opérateur de division
 * Coords<T, N> Coords<T, N>::operator /(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorDivision, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		try
		{
			Coords<T, 2> c23(c21 / c22);
			for (size_t i = 0; i < c23.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) / dataAccess(c22, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			clog << "infinite component in /" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			Coords<T, 3> c33(c31 / c32);

			for (size_t i = 0; i < c33.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) / dataAccess(c32, i)),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			clog << "infinite component in /" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			Coords<T, 4> c43(c41 / c42);
			for (size_t i = 0; i < c43.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) / dataAccess(c42, i)),
				                  tol);
			}

		}
		catch (CoordsException & e)
		{
			clog << "infinite component in /" << endl;

			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
	}

	/*
	 * Checking infinite values
	 */

	T infiniteValue = numeric_limits<T>::max();
	T nullValue = numeric_limits<T>::min();
	Coords<T, 2> c21(infiniteValue, infiniteValue);
	Coords<T, 3> c31(infiniteValue, infiniteValue, infiniteValue);
	Coords<T, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

	Coords<T, 2> c22(nullValue, nullValue);
	Coords<T, 3> c32(nullValue, nullValue, nullValue);
	Coords<T, 4> c42(nullValue, nullValue, nullValue, nullValue);

	try
	{
		Coords<T, 2> c23(c21 / c22);

		for (size_t i = 0; i < c23.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c23, i)),
			                  static_cast<double>(dataAccess(c21, i) / dataAccess(c22, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in /" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
	try
	{
		Coords<T, 3> c33(c31 /c32);

		for (size_t i = 0; i < c33.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c33, i)),
			                  static_cast<double>(dataAccess(c31, i) / dataAccess(c32, i)),
			                  tol);
		}
	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in /" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}

	try
	{
		Coords<T, 4> c43(c41 / c42);

		for (size_t i = 0; i < c43.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c43, i)),
			                  static_cast<double>(dataAccess(c41, i) / dataAccess(c42, i)),
			                  tol);
		}

	}
	catch (CoordsException & e)
	{
		// cerr << "infinite component in /" << endl;
		BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
	}
}

/**
 * Test de la méthode dotProduct
 * T Coords<T, N>::dotProduct(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( DotProduct, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		Coords<T, 2> c22(random(), random());
		Coords<T, 3> c32(random(), random(), random());
		Coords<T, 4> c42(random(), random(), random(), random());

		T d2 = c21.dotProduct(c22);
		T d3 = c31.dotProduct(c32);
		T d4 = c41.dotProduct(c42);

		T sum2 = static_cast<T>(0);
		for (size_t i = 0; i < c21.size(); i++)
		{
			sum2 += dataAccess(c21, i) * dataAccess(c22, i);
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d2), static_cast<double>(sum2));

		T sum3 = static_cast<T>(0);
		for (size_t i = 0; i < c31.size(); i++)
		{
			sum3 += dataAccess(c31, i) * dataAccess(c32, i);
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d3), static_cast<double>(sum3));

		T sum4 = static_cast<T>(0);
		for (size_t i = 0; i < c41.size(); i++)
		{
			sum4 += dataAccess(c41, i) * dataAccess(c42, i);
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d4), static_cast<double>(sum4));
	}
}

/**
 * Test de la méthode norm2
 * T Coords<T, N>::norm2() const
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Norm2, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c21(random(), random());
		Coords<T, 3> c31(random(), random(), random());
		Coords<T, 4> c41(random(), random(), random(), random());

		T d2 = c21.norm2();
		T d3 = c31.norm2();
		T d4 = c41.norm2();

		T sum2 = static_cast<T>(0);
		for (size_t i = 0; i < c21.size(); i++)
		{
			T val = dataAccess(c21, i);
			sum2 += val * val;
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d2), static_cast<double>(sum2));

		T sum3 = static_cast<T>(0);
		for (size_t i = 0; i < c31.size(); i++)
		{
			T val = dataAccess(c31, i);
			sum3 += val * val;
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d3), static_cast<double>(sum3));

		T sum4 = static_cast<T>(0);
		for (size_t i = 0; i < c41.size(); i++)
		{
			T val = dataAccess(c41, i);
			sum4 += val * val;
		}

		BOOST_CHECK_EQUAL(static_cast<double>(d4), static_cast<double>(sum4));
	}
}

/**
 * Test de la méthode translate
 * Coords<T, N> & Coords<T, N>::tranlate(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Translate, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c2(random(), random());
		Coords<T, 3> c3(random(), random(), random());
		Coords<T, 4> c4(random(), random(), random(), random());

		Coords<T, 2> c2Saved(c2);
		Coords<T, 3> c3Saved(c3);
		Coords<T, 4> c4Saved(c4);

		Coords<T, 2> t21(random(), random());
		Coords<T, 3> t31(random(), random(), random());
		Coords<T, 4> t41(random(), random(), random(), random());
		Coords<T, 2> t22(random(), random());
		Coords<T, 3> t32(random(), random(), random());
		Coords<T, 4> t42(random(), random(), random(), random());

		c2.translate(t21).translate(t22);
		c3.translate(t31).translate(t32);
		c4.translate(t41).translate(t42);

		for (size_t i = 0; i < c2.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c2, i)),
			                  static_cast<double>(dataAccess(c2Saved, i) +
			                                      dataAccess(t21, i) +
			                                      dataAccess(t22, i)),
			                  tol);
		}

		for (size_t i = 0; i < c3.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c3, i)),
			                  static_cast<double>(dataAccess(c3Saved, i) +
			                                      dataAccess(t31, i) +
			                                      dataAccess(t32, i)),
			                  tol);
		}

		for (size_t i = 0; i < c4.size(); i++)
		{
			BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c4, i)),
			                  static_cast<double>(dataAccess(c4Saved, i) +
			                                      dataAccess(t41, i) +
			                                      dataAccess(t42, i)),
			                  tol);
		}
	}
}

/**
 * Test de la méthode scale
 * Coords<T, N> & Coords<T, N>::tranlate(Const<T, N> &)
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( Scale, T)
{
	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c2(random(), random());
		Coords<T, 3> c3(random(), random(), random());
		Coords<T, 4> c4(random(), random(), random(), random());

		Coords<T, 2> c2Saved(c2);
		Coords<T, 3> c3Saved(c3);
		Coords<T, 4> c4Saved(c4);

		T s1(random());
		T s2(random());

		try
		{
			c2.scale(s1).scale(s2);

			for (size_t i = 0; i < c2.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c2, i)),
				                  static_cast<double>(dataAccess(c2Saved, i) *
				                                      s1 * s2),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c3.scale(s1).scale(s2);

			for (size_t i = 0; i < c3.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c3, i)),
				                  static_cast<double>(dataAccess(c3Saved, i) *
				                                      s1 * s2),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c4.scale(s1).scale(s2);
			for (size_t i = 0; i < c4.size(); i++)
			{
				BOOST_CHECK_CLOSE(static_cast<double>(dataAccess(c4, i)),
				                  static_cast<double>(dataAccess(c4Saved, i) *
				                                      s1 * s2),
				                  tol);
			}
		}
		catch (CoordsException & e)
		{
			BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		/*
		 * Checking infinite values
		 */
		if (numeric_limits<T>::has_infinity)
		{
			T sinf(numeric_limits<T>::max());
			try
			{
				c2.scale(sinf);

				BOOST_FAIL("Passed Infinite Scale");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c3.scale(sinf);

				BOOST_FAIL("Passed Infinite Scale");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c4.scale(sinf);

				BOOST_FAIL("Passed Infinite Scale");
			}
			catch (CoordsException & e)
			{
				BOOST_CHECK(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}
	}
}

/**
 * Test de l'opérateur de sortie standard
 * operator <<(ostream &, const Coords<T> &);
 */
BOOST_TEST_CASE_TEMPLATE_FUNCTION( OperatorOutput, T )
{
	output_test_stream output;

	function<T()> random = getRandomGenerator<T>();

	for (size_t t = 0; t < nbNumericTests; t++)
	{
		Coords<T, 2> c2(random(), random());

		output << c2;

		/*
		 * Le flux de sortie n'est pas vide
		 */
		BOOST_CHECK(!output.is_empty(false));

		/*
		 * La chaîne attendue pour une coordonnée est :
		 * (x1, x2,    , xn)
		 */
		ostringstream oss2;
		oss2 << '(';

		size_t nbCoords = Coords<T, 2>::size();
		for (size_t i = 0; i < nbCoords; i++)
		{
			oss2 << dataAccess(c2, i);
			if (i < (nbCoords - 1))
			{
				oss2 << ", ";
			}
		}

		oss2 << ')';

		string expected = oss2.str();

		/*
		 * Le flux de sortie contient la chaine attendue
		 */
		BOOST_CHECK(output.is_equal(expected));

		output.flush();
		expected.clear();

		Coords<T, 3> c3(random(), random(), random());

		output << c3;

		/*
		 * Le flux de sortie n'est pas vide
		 */
		BOOST_CHECK(!output.is_empty(false));

		/*
		 * La chaîne attendue pour une coordonnée est :
		 * (x1, x2,    , xn)
		 */
		ostringstream oss3;
		oss3 << '(';

		nbCoords = Coords<T, 3>::size();
		for (size_t i = 0; i < nbCoords; i++)
		{
			oss3 << dataAccess(c3, i);
			if (i < (nbCoords - 1))
			{
				oss3 << ", ";
			}
		}

		oss3 << ')';

		expected = oss3.str();

		/*
		 * Le flux de sortie contient la chaine attendue
		 */
		BOOST_CHECK(output.is_equal(expected));

		output.flush();
		expected.clear();

		Coords<T, 4> c4(random(), random(), random(), random());

		output << c4;

		/*
		 * Le flux de sortie n'est pas vide
		 */
		BOOST_CHECK(!output.is_empty(false));

		/*
		 * La chaîne attendue pour une coordonnée est :
		 * (x1, x2,    , xn)
		 */
		ostringstream oss4;
		oss4 << '(';

		nbCoords = Coords<T, 4>::size();
		for (size_t i = 0; i < nbCoords; i++)
		{
			oss4 << dataAccess(c4, i);
			if (i < (nbCoords - 1))
			{
				oss4 << ", ";
			}
		}

		oss4 << ')';

		expected = oss4.str();

		/*
		 * Le flux de sortie contient la chaine attendue
		 */
		BOOST_CHECK(output.is_equal(expected));

		output.flush();
	}
}

// ----------------------------------------------------------------------------
// Suites de tests
// ----------------------------------------------------------------------------
/**
 * Initialisation de la suite de tests
 * @return 0
 */
test_suite* init_unit_test_suite(int, char**)
{
	typedef boost::mpl::list<double, float, int> test_types;

	// Coords<T> test set
	test_suite * cts = BOOST_TEST_SUITE( "Coords<T>" );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Coords_, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Coords_Variadic, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Coords_Initializer_List, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(CoordsOfCoords, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(CoordsOfCoordsUM, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorCopyOfCoords, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(GetConst, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Get, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorBracketConst, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorBracket, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorCompare, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorDifference, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorSelfAddition, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorSelfSubstraction, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorSelfMultiplication, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorSelfDivision, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorAddition, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorSubstraction, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorMultiplication, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorDivision, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(DotProduct, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Norm2, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Translate, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(Scale, test_types) );
	cts->add( BOOST_TEST_CASE_TEMPLATE(OperatorOutput, test_types) );

	framework::master_test_suite().add(cts);

	return 0;
}
