/*
 * main.cpp
 *
 *  Created on: 30 juin 2011
 *      Author: davidroussel
 */

#include <cstdlib>
#include <iostream>
#include <exception>
#include <sstream>		// for ostringstream
#include <cassert>
#include <random>	// for random numbers dist. & generator
#include <limits>	// for numeric_limits
#include <functional>
using namespace std;

#include "Coords.h"
#include "CoordsException.h"

/**
 * Creates a random number generator with values contained within
 * [minVal ... maxVal]
 * @tparam T the type of values to generate
 * @param minVal the minimum value
 * @param maxVal the maximum value
 * @return a function returning the next random number
 */
template<typename T>
function<T(void)> getRandomGenerator(const T & minVal = static_cast<T>(-2147483647),
                                     const T & maxVal = static_cast<T>(2147483648))
{
	// clog << "Generating values between" << minVal << " and " << maxVal << endl;
	default_random_engine engine;
	function<T(void)> random;
	if (typeid(T) == typeid(size_t))
	{
		random = bind(uniform_int_distribution<size_t>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(int))
	{
		random = bind(uniform_int_distribution<int>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(float))
	{
		random = bind(uniform_real_distribution<float>(minVal, maxVal), engine);
	}
	else if (typeid(T) == typeid(double))
	{
		random = bind(uniform_real_distribution<double>(minVal, maxVal), engine);
	}
	else
	{
		cerr << "getRandomGenerator() : Unknown value type" << typeid(T).name()
		     << endl;
	}

	return random;
}

/**
 * Compare two values equality within a tolerance expressed in percent
 * @param val1 the first value to compare
 * @param val2 the second value to compare
 * @param relTol relative tolerance for comparing two values
 * @param absTol Absolute tolerance for comparing two values
 * @return true if val1 == val2 evt within relative tolerance or absolute
 * tolerance when relative tolerance is not applicable (around 0)
 */
template<typename T>
bool checkClose(const T & val1,
                const T & val2,
                const double relTol,
                const T & absTol)
{
	if (val1 == val2)
	{
//		clog << "Comparing " << val1 << " == " << val2 << endl;
		return true;
	}

	double val1d = double(val1);
	double val2d = double(val2);

	double absoluteError = abs(val1d - val2d);

//	clog << "Comparing(" << val1 << ", " << val2 << ") absolute error = "
//	     << absoluteError << (absoluteError < absTol ? " < " : " >= ")
//	     << absTol << endl;

	if (absoluteError < double(absTol))
	{
		return true;
	}

	double relativeError;
	if (abs(val2d) > abs(val1d))
	{
		relativeError = abs((val1d - val2d) / val2d);
	}
	else
	{
		relativeError = abs((val1d - val2d) / val1d);
	}

//	clog << "Comparing(" << val1 << ", " << val2 << ") relative error = "
//	     << relativeError << (relativeError < relTol ? " < " : " >= ")
//	     << relTol << endl;

	if (relativeError <= relTol)
	{
		return true;
	}

	return false;
}

int main(void)
{
	size_t nbNumericTests = 100;
	double relTol = 0.0001; // Tolerance Percentage when checking numeric value
	double absTol = 0.0000001;
	function<double()> random = getRandomGenerator<double>();

	// ------------------------------------------------------------------------
	cout << "Test de static size_t size() ... ";
	{
		Coords<double, 2> c2;
		assert(c2.size() == 2);

		Coords<double, 3> c3;
		assert(c3.size() == 3);

		Coords<double, 4> c4;
		assert(c4.size() == 4);
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords() ... ";
	{
		Coords<double, 2> c2;

		for (size_t i = 0; i < c2.size(); i++)
		{
			assert(dataAccess(c2, i) == 0.0);
		}

		Coords<double, 3> c3;

		for (size_t i = 0; i < c3.size(); i++)
		{
			assert(dataAccess(c3, i) == 0.0);
		}

		Coords<double, 4> c4;

		for (size_t i = 0; i < c4.size(); i++)
		{
			assert(dataAccess(c4, i) == 0.0);
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords(const T & value, const U & ... others) ... ";
	{
		double values[] = {random(), random(), random(), random()};

		Coords<double, 2> c21(values[0]);
		Coords<double, 2> c22(values[0], values[1]);
		Coords<double, 2> c23(values[0], values[1], values[2]);
		Coords<double, 2> c24(values[0], values[1], values[2], values[3]);
		Coords<double, 2> * tc2[] = {&c21, &c22, &c23, &c24};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc2[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc2[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc2[c]), i) == values[i]);
				}
			}
		}

		Coords<double, 3> c31(values[0]);
		Coords<double, 3> c32(values[0], values[1]);
		Coords<double, 3> c33(values[0], values[1], values[2]);
		Coords<double, 3> c34(values[0], values[1], values[2], values[3]);
		Coords<double, 3> * tc3[] = {&c31, &c32, &c33, &c34};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc3[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc3[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc3[c]), i) == values[i]);
				}
			}
		}

		Coords<double, 4> c41(values[0]);
		Coords<double, 4> c42(values[0], values[1]);
		Coords<double, 4> c43(values[0], values[1], values[2]);
		Coords<double, 4> c44(values[0], values[1], values[2], values[3]);
		Coords<double, 4> * tc4[] = {&c41, &c42, &c43, &c44};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc4[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc4[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc4[c]), i) == values[i]);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords(std::initializer_list<T> list) ... ";
	{
		double values[] = {random(), random(), random(), random()};

		Coords<double, 2> c21 = {values[0]};
		Coords<double, 2> c22 = {values[0], values[1]};
		Coords<double, 2> c23 = {values[0], values[1], values[2]};
		Coords<double, 2> c24 = {values[0], values[1], values[2], values[3]};
		Coords<double, 2> * tc2[] = {&c21, &c22, &c23, &c24};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc2[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc2[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc2[c]), i) == values[i]);
				}
			}
		}

		Coords<double, 3> c31 = {values[0]};
		Coords<double, 3> c32 = {values[0], values[1]};
		Coords<double, 3> c33 = {values[0], values[1], values[2]};
		Coords<double, 3> c34 = {values[0], values[1], values[2], values[3]};
		Coords<double, 3> * tc3[] = {&c31, &c32, &c33, &c34};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc3[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc3[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc3[c]), i) == values[i]);
				}
			}
		}

		Coords<double, 4> c41 = {values[0]};
		Coords<double, 4> c42 = {values[0], values[1]};
		Coords<double, 4> c43 = {values[0], values[1], values[2]};
		Coords<double, 4> c44 = {values[0], values[1], values[2], values[3]};
		Coords<double, 4> * tc4[] = { &c41, &c42, &c43, &c44};

		for (size_t c = 0; c < 4; c++)
		{
			for (size_t i = 0; i < tc4[c]->size(); i++)
			{
				if (i > c)
				{
					assert(dataAccess(*(tc4[c]), i) == 0.0);
				}
				else
				{
					assert(dataAccess(*(tc4[c]), i) == values[i]);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords(const Coords<T, N> & c) ... ";
	{
		double values[] = {random(), random(), random(), random()};

		Coords<double, 2> c21(values[0], values[1], values[2], values[3]);
		Coords<double, 2> c22(c21);
		for (size_t i = 0; i < c22.size(); i++)
		{
			assert(dataAccess(c22, i) == dataAccess(c21, i));
		}

		Coords<double, 2> c23 = c21;
		for (size_t i = 0; i < c23.size(); i++)
		{
			assert(dataAccess(c23, i) == dataAccess(c22, i));
		}

		Coords<double, 3> c31(values[0], values[1], values[2], values[3]);
		Coords<double, 3> c32(c31);
		for (size_t i = 0; i < c32.size(); i++)
		{
			assert(dataAccess(c32, i) == dataAccess(c31, i));
		}

		Coords<double, 3> c33 = c31;
		for (size_t i = 0; i < c33.size(); i++)
		{
			assert(dataAccess(c33, i) == dataAccess(c32, i));
		}

		Coords<double, 4> c41(values[0], values[1], values[2], values[3]);
		Coords<double, 4> c42(c41);
		for (size_t i = 0; i < c42.size(); i++)
		{
			assert(dataAccess(c42, i) == dataAccess(c41, i));
		}

		Coords<double, 4> c43 = c41;
		for (size_t i = 0; i < c43.size(); i++)
		{
			assert(dataAccess(c43, i) == dataAccess(c42, i));
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords(const Coords<U, M> & c) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			double values[] = {random(), random(), random(), random()};

			Coords<double, 2> c2(values[0], values[1]);
			Coords<double, 3> c3(values[0], values[1], values[2]);
			Coords<double, 4> c4(values[0], values[1], values[2], values[3]);

			Coords<double, 2> cd22(c2);
			Coords<double, 3> cd32(c2);
			Coords<double, 4> cd42(c2);
			Coords<float, 2> cf22(c2);
			Coords<float, 3> cf32(c2);
			Coords<float, 4> cf42(c2);
			Coords<int, 2> ci22(c2);
			Coords<int, 3> ci32(c2);
			Coords<int, 4> ci42(c2);

			Coords<double, 2> cd23(c3);
			Coords<double, 3> cd33(c3);
			Coords<double, 4> cd43(c3);
			Coords<float, 2> cf23(c3);
			Coords<float, 3> cf33(c3);
			Coords<float, 4> cf43(c3);
			Coords<int, 2> ci23(c3);
			Coords<int, 3> ci33(c3);
			Coords<int, 4> ci43(c3);

			Coords<double, 2> cd24(c4);
			Coords<double, 3> cd34(c4);
			Coords<double, 4> cd44(c4);
			Coords<float, 2> cf24(c4);
			Coords<float, 3> cf34(c4);
			Coords<float, 4> cf44(c4);
			Coords<int, 2> ci24(c4);
			Coords<int, 3> ci34(c4);
			Coords<int, 4> ci44(c4);

			/*
			 * Check values from c2, c3, c4
			 */
			for (size_t i = 0; i < c2.size(); i++)
			{
				assert(static_cast<double>(dataAccess(c2, i)) == dataAccess(cd22, i));
				assert(static_cast<double>(dataAccess(c2, i)) == dataAccess(cd32, i));
				assert(static_cast<double>(dataAccess(c2, i)) == dataAccess(cd42, i));
				assert(static_cast<float>(dataAccess(c2, i)) == dataAccess(cf22, i));
				assert(static_cast<float>(dataAccess(c2, i)) == dataAccess(cf32, i));
				assert(static_cast<float>(dataAccess(c2, i)) == dataAccess(cf42, i));
				assert(static_cast<int>(dataAccess(c2, i)) == dataAccess(ci22, i));
				assert(static_cast<int>(dataAccess(c2, i)) == dataAccess(ci32, i));
				assert(static_cast<int>(dataAccess(c2, i)) == dataAccess(ci42, i));

				assert(static_cast<double>(dataAccess(c3, i)) == dataAccess(cd23, i));
				assert(static_cast<double>(dataAccess(c3, i)) == dataAccess(cd33, i));
				assert(static_cast<double>(dataAccess(c3, i)) == dataAccess(cd43, i));
				assert(static_cast<float>(dataAccess(c3, i)) == dataAccess(cf23, i));
				assert(static_cast<float>(dataAccess(c3, i)) == dataAccess(cf33, i));
				assert(static_cast<float>(dataAccess(c3, i)) == dataAccess(cf43, i));
				assert(static_cast<int>(dataAccess(c3, i)) == dataAccess(ci23, i));
				assert(static_cast<int>(dataAccess(c3, i)) == dataAccess(ci33, i));
				assert(static_cast<int>(dataAccess(c3, i)) == dataAccess(ci43, i));

				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd24, i));
				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd34, i));
				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd44, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf24, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf34, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf44, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci24, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci34, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci44, i));
			}

			for (size_t i = c2.size(); i < c3.size(); i++)
			{
				assert(static_cast<double>(dataAccess(c3, i)) == dataAccess(cd33, i));
				assert(static_cast<double>(dataAccess(c3, i)) == dataAccess(cd43, i));
				assert(static_cast<float>(dataAccess(c3, i)) == dataAccess(cf33, i));
				assert(static_cast<float>(dataAccess(c3, i)) == dataAccess(cf43, i));
				assert(static_cast<int>(dataAccess(c3, i)) == dataAccess(ci33, i));
				assert(static_cast<int>(dataAccess(c3, i)) == dataAccess(ci43, i));

				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd34, i));
				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd44, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf34, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf44, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci34, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci44, i));

				assert(dataAccess(cd32, i) == 0.0);
				assert(dataAccess(cd42, i) == 0.0);
				assert(dataAccess(cf32, i) == 0.0f);
				assert(dataAccess(cf42, i) == 0.0f);
				assert(dataAccess(ci32, i) == 0);
				assert(dataAccess(ci42, i) == 0);

			}

			for (size_t i = c3.size(); i < c4.size(); i++)
			{
				assert(static_cast<double>(dataAccess(c4, i)) == dataAccess(cd44, i));
				assert(static_cast<float>(dataAccess(c4, i)) == dataAccess(cf44, i));
				assert(static_cast<int>(dataAccess(c4, i)) == dataAccess(ci44, i));

				assert(dataAccess(cd42, i) == 0.0);
				assert(dataAccess(cd43, i) == 0.0);
				assert(dataAccess(cf42, i) == 0.0f);
				assert(dataAccess(cf43, i) == 0.0f);
				assert(dataAccess(ci42, i) == 0);
				assert(dataAccess(ci43, i) == 0);
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T get(const size_t i) const throw (CoordsException) ... ";
	{
		function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

		for (size_t t = 0; t < nbNumericTests; t++)
		{
			double values[] = {random(), random(), random(), random()};

			Coords<double, 2> c21(values[0], values[1]);
			Coords<double, 3> c31(values[0], values[1], values[2]);
			Coords<double, 4> c41(values[0], values[1], values[2], values[3]);

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(c21.get(i) == values[i]);
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(c31.get(i) == values[i]);
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(c41.get(i) == values[i]);
			}

			/*
			 * We should get an CoordsException if we try to get invalid index
			 */
			size_t i = indexRandom();
			if (i >= c21.size())
			{
				try
				{
					double value = c21.get(i);

					cerr << "Getting Value for index " << i << " in " << c21  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c31.size())
			{
				try
				{
					double value = c31.get(i);

					cerr << "Getting Value for index " << i << " in " << c31  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c41.size())
			{
				try
				{
					double value = c41.get(i);

					cerr << "Getting Value for index " << i << " in " << c41  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T & get(const size_t i) throw (CoordsException) ... ";
	{
		auto indexRandom = getRandomGenerator<size_t>();

		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21;
			Coords<double, 3> c31;
			Coords<double, 4> c41;

			for (size_t i = 0; i < c21.size(); i++)
			{
				double value = random();
				c21.get(i) = value;
				assert(dataAccess(c21, i) == value);
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				double value = random();
				c31.get(i) = value;
				assert(dataAccess(c31, i) == value);
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				double value = random();
				c41.get(i) = value;
				assert(dataAccess(c41, i) == value);
			}

			/*
			 * We should get an CoordsException if we try to get invalid index
			 */
			size_t i = indexRandom();
			if (i >= c21.size())
			{
				try
				{
					double value = random();
					c21.get(i) = value;

					cerr << "Getting Value for index " << i << " in " << c21  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c31.size())
			{
				try
				{
					double value = random();
					c31.get(i) = value;

					cerr << "Getting Value for index " << i << " in " << c31  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c41.size())
			{
				try
				{
					double value = random();
					c41.get(i) = value;

					cerr << "Getting Value for index " << i << " in " << c41  << " = " << value << "succeded";
					return EXIT_FAILURE;
				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T operator [](const size_t i) const throw (CoordsException) ...";
	{
		function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

		for (size_t t = 0; t < nbNumericTests; t++)
		{
			double values[] = {random(), random(), random(), random()};

			Coords<double, 2> c21(values[0], values[1]);
			Coords<double, 3> c31(values[0], values[1], values[2]);
			Coords<double, 4> c41(values[0], values[1], values[2], values[3]);

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(c21[i] == values[i]);
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(c31[i] == values[i]);
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(c41[i] == values[i]);
			}

			/*
			 * We should get an CoordsException if we try to get invalid index
			 */
			size_t i = indexRandom();
			if (i >= c21.size())
			{
				try
				{
					double value = c21[i];

					cerr << "Getting Value for index " << i << " in " << c21  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c31.size())
			{
				try
				{
					double value = c31[i];

					cerr << "Getting Value for index " << i << " in " << c31  << " = " << value << "succeded";
					return EXIT_FAILURE;

				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c41.size())
			{
				try
				{
					double value = c41[i];

					cerr << "Getting Value for index " << i << " in " << c41  << " = " << value << "succeded";
					return EXIT_FAILURE;
				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T & operator [](const size_t i) throw (CoordsException) ... ";
	{
		function<size_t(void)> indexRandom = getRandomGenerator<size_t>();

		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21;
			Coords<double, 3> c31;
			Coords<double, 4> c41;

			for (size_t i = 0; i < c21.size(); i++)
			{
				double value = random();
				c21[i] = value;
				assert(dataAccess(c21, i) == value);
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				double value = random();
				c31[i] = value;
				assert(dataAccess(c31, i) == value);
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				double value = random();
				c41[i] = value;
				assert(dataAccess(c41, i) == value);
			}

			/*
			 * We should get an CoordsException if we try to get invalid index
			 */
			size_t i = indexRandom();
			if (i >= c21.size())
			{
				try
				{
					double value = random();
					c21[i] = value;

					cerr << "Getting Value for index " << i << " in " << c21  << " = " << value << "succeded";
					return EXIT_FAILURE;
				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c31.size())
			{
				try
				{
					double value = random();
					c31[i] = value;

					cerr << "Getting Value for index " << i << " in " << c31  << " = " << value << "succeded";
					return EXIT_FAILURE;
				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
			if (i >= c41.size())
			{
				try
				{
					double value = random();
					c41[i] = value;

					cerr << "Getting Value for index " << i << " in " << c41  << " = " << value << "succeded";
					return EXIT_FAILURE;
				}
				catch (CoordsException & e)
				{
					assert(e.getCause() == CoordsException::ExceptionCause::INDICES_OUT_OF_BOUNDS);
				}
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & operator =(const Coords<U, M> & c) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22;
			Coords<double, 3> c32;
			Coords<double, 4> c42;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(dataAccess(c21, i) != dataAccess(c22, i));
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(dataAccess(c31, i) != dataAccess(c32, i));
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(dataAccess(c41, i) != dataAccess(c42, i));
			}

			c22 = c21;
			c32 = c31;
			c42 = c41;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(dataAccess(c21, i) == dataAccess(c22, i));
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(dataAccess(c31, i) == dataAccess(c32, i));
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(dataAccess(c41, i) == dataAccess(c42, i));
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de bool operator ==(const Coords<T, N> & c) const ...";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22;
			Coords<double, 3> c32;
			Coords<double, 4> c42;

			assert(!(c21 == c22));

			assert(!(c31 == c32));

			assert(!(c41 == c42));

			c22 = c21;
			c32 = c31;
			c42 = c41;

			assert(c21 == c22);

			assert(c31 == c32);

			assert(c41 == c42);
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de bool operator !=(const Coords<T, N> & c) const ...";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22;
			Coords<double, 3> c32;
			Coords<double, 4> c42;

			assert(c21 != c22);

			assert(c31 != c32);

			assert(c41 != c42);

			c22 = c21;
			c32 = c31;
			c42 = c41;

			assert(!(c21 != c22));

			assert(!(c31 != c32));

			assert(!(c41 != c42));
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & operator +=(const Coords<T, N> & c) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c21Saved(c21);
			Coords<double, 3> c31Saved(c31);
			Coords<double, 4> c41Saved(c41);

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			c21 += c22;
			c31 += c32;
			c41 += c42;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(checkClose<double>(static_cast<double>(dataAccess(c21, i)),
				                          static_cast<double>(dataAccess(c21Saved, i) +
				                                              dataAccess(c22, i)),
				                          relTol,
				                          absTol));
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) +
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) +
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & operator -=(const Coords<T, N> & c) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c21Saved(c21);
			Coords<double, 3> c31Saved(c31);
			Coords<double, 4> c41Saved(c41);

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			c21 -= c22;
			c31 -= c32;
			c41 -= c42;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c21, i)),
				                  static_cast<double>(dataAccess(c21Saved, i) -
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) -
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) -
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & operator *=(const Coords<T, N> & c) throw (CoordsException) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c21Saved(c21);
			Coords<double, 3> c31Saved(c31);
			Coords<double, 4> c41Saved(c41);

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			try
			{
				c21 *= c22;

				for (size_t i = 0; i < c21.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c21, i)),
					                  static_cast<double>(dataAccess(c21Saved, i) *
					                                      dataAccess(c22, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// cerr << c21 << " has infinite component in *=" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
			try
			{
				c31 *= c32;

				for (size_t i = 0; i < c31.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c31, i)),
					                  static_cast<double>(dataAccess(c31Saved, i) *
					                                      dataAccess(c32, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// cerr << c31 << " has infinite component in *=" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c41 *= c42;

				for (size_t i = 0; i < c41.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c41, i)),
					                  static_cast<double>(dataAccess(c41Saved, i) *
					                                      dataAccess(c42, i)),
					                  relTol,
					                  absTol));
				}

			}
			catch (CoordsException & e)
			{
				// cerr << c41 << " has infinite component in *=" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}

		/*
		 * Checking infinite values
		 */

		double infiniteValue = numeric_limits<double>::max();
		Coords<double, 2> c21(infiniteValue, infiniteValue);
		Coords<double, 3> c31(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		Coords<double, 2> c21Saved(c21);
		Coords<double, 3> c31Saved(c31);
		Coords<double, 4> c41Saved(c41);

		Coords<double, 2> c22(infiniteValue, infiniteValue);
		Coords<double, 3> c32(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c42(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		try
		{
			c21 *= c22;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c21, i)),
				                  static_cast<double>(dataAccess(c21Saved, i) *
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c21 << " has infinite component in *=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			c31 *= c32;

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) *
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c31 << " has infinite component in *=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c41 *= c42;

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) *
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}

		}
		catch (CoordsException & e)
		{
			// cerr << c41 << " has infinite component in *=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & operator /=(const Coords<T, N> & c) throw (CoordsException) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c21Saved(c21);
			Coords<double, 3> c31Saved(c31);
			Coords<double, 4> c41Saved(c41);

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			try
			{
				c21 /= c22;
				for (size_t i = 0; i < c21.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c21, i)),
					                  static_cast<double>(dataAccess(c21Saved, i) /
					                                      dataAccess(c22, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// clog << c21 << " has infinite component in /=" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
			try
			{
				c31 /= c32;

				for (size_t i = 0; i < c31.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c31, i)),
					                  static_cast<double>(dataAccess(c31Saved, i) /
					                                      dataAccess(c32, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// clog << c31 << " has infinite component in /=" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c41 /= c42;
				for (size_t i = 0; i < c41.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c41, i)),
					                  static_cast<double>(dataAccess(c41Saved, i) /
					                                      dataAccess(c42, i)),
					                  relTol,
					                  absTol));
				}

			}
			catch (CoordsException & e)
			{
				// clog << c41 << " has infinite component in /=" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}

		/*
		 * Checking infinite values
		 */

		double infiniteValue = numeric_limits<double>::max();
		double nullValue = numeric_limits<double>::min();
		Coords<double, 2> c21(infiniteValue, infiniteValue);
		Coords<double, 3> c31(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		Coords<double, 2> c21Saved(c21);
		Coords<double, 3> c31Saved(c31);
		Coords<double, 4> c41Saved(c41);

		Coords<double, 2> c22(nullValue, nullValue);
		Coords<double, 3> c32(nullValue, nullValue, nullValue);
		Coords<double, 4> c42(nullValue, nullValue, nullValue, nullValue);

		try
		{
			c21 /= c22;

			for (size_t i = 0; i < c21.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c21, i)),
				                  static_cast<double>(dataAccess(c21Saved, i) /
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c21 << " has infinite component in /=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			c31 /= c32;

			for (size_t i = 0; i < c31.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c31, i)),
				                  static_cast<double>(dataAccess(c31Saved, i) /
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << c31 << " has infinite component in /=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			c41 /= c42;

			for (size_t i = 0; i < c41.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c41, i)),
				                  static_cast<double>(dataAccess(c41Saved, i) /
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}

		}
		catch (CoordsException & e)
		{
			// cerr << c41 << " has infinite component in /=" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> operator +(const Coords<T, N> & c) const ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			Coords<double, 2> c23(c21 + c22);
			Coords<double, 2> c33(c31 + c32);
			Coords<double, 2> c43(c41 + c42);

			for (size_t i = 0; i < c23.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) +
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c33.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) +
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c43.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) +
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> operator -(const Coords<T, N> & c) const ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			Coords<double, 2> c23(c21 - c22);
			Coords<double, 2> c33(c31 - c32);
			Coords<double, 2> c43(c41 - c42);

			for (size_t i = 0; i < c23.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) -
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c33.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) -
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c43.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) -
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> operator *(const Coords<T, N> & c) const throw (CoordsException) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			try
			{
				Coords<double, 2> c23(c21 * c22);
				for (size_t i = 0; i < c21.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c23, i)),
					                  static_cast<double>(dataAccess(c21, i) *
					                                      dataAccess(c22, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// cerr << "infinite component in *" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
			try
			{
				Coords<double, 3> c33(c31 * c32);

				for (size_t i = 0; i < c31.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c33, i)),
					                  static_cast<double>(dataAccess(c31, i) *
					                                      dataAccess(c32, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				// cerr << "infinite component in *" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				Coords<double, 4> c43(c41 * c42);

				for (size_t i = 0; i < c41.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c43, i)),
					                  static_cast<double>(dataAccess(c41, i) *
					                                      dataAccess(c42, i)),
					                  relTol,
					                  absTol));
				}

			}
			catch (CoordsException & e)
			{
				// cerr << "infinite component in *" << endl;
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}

		/*
		 * Checking infinite values
		 */

		double infiniteValue = numeric_limits<double>::max();
		Coords<double, 2> c21(infiniteValue, infiniteValue);
		Coords<double, 3> c31(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		Coords<double, 2> c22(infiniteValue, infiniteValue);
		Coords<double, 3> c32(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c42(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		try
		{
			Coords<double, 2> c23(c21 *c22);

			for (size_t i = 0; i < c23.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) * dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			Coords<double, 3> c33(c31 * c32);

			for (size_t i = 0; i < c33.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) *
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			Coords<double, 4> c43(c41 *c42);

			for (size_t i = 0; i < c43.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) *
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}

		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in *" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> operator /(const Coords<T, N> & c) const throw (CoordsException) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			try
			{
				Coords<double, 2> c23(c21 / c22);
				for (size_t i = 0; i < c23.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c23, i)),
					                  static_cast<double>(dataAccess(c21, i) /
					                                      dataAccess(c22, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				clog << "infinite component in /" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
			try
			{
				Coords<double, 3> c33(c31 / c32);

				for (size_t i = 0; i < c33.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c33, i)),
					                  static_cast<double>(dataAccess(c31, i) /
					                                      dataAccess(c32, i)),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				clog << "infinite component in /" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				Coords<double, 4> c43(c41 / c42);
				for (size_t i = 0; i < c43.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c43, i)),
					                  static_cast<double>(dataAccess(c41, i) /
					                                      dataAccess(c42, i)),
					                  relTol,
					                  absTol));
				}

			}
			catch (CoordsException & e)
			{
				clog << "infinite component in /" << endl;

				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}

		/*
		 * Checking infinite values
		 */

		double infiniteValue = numeric_limits<double>::max();
		double nullValue = numeric_limits<double>::min();
		Coords<double, 2> c21(infiniteValue, infiniteValue);
		Coords<double, 3> c31(infiniteValue, infiniteValue, infiniteValue);
		Coords<double, 4> c41(infiniteValue, infiniteValue, infiniteValue, infiniteValue);

		Coords<double, 2> c22(nullValue, nullValue);
		Coords<double, 3> c32(nullValue, nullValue, nullValue);
		Coords<double, 4> c42(nullValue, nullValue, nullValue, nullValue);

		try
		{
			Coords<double, 2> c23(c21 / c22);

			for (size_t i = 0; i < c23.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c23, i)),
				                  static_cast<double>(dataAccess(c21, i) /
				                                      dataAccess(c22, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in /" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
		try
		{
			Coords<double, 3> c33(c31 /c32);

			for (size_t i = 0; i < c33.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c33, i)),
				                  static_cast<double>(dataAccess(c31, i) /
				                                      dataAccess(c32, i)),
				                  relTol,
				                  absTol));
			}
		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in /" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}

		try
		{
			Coords<double, 4> c43(c41 / c42);

			for (size_t i = 0; i < c43.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c43, i)),
				                  static_cast<double>(dataAccess(c41, i) /
				                                      dataAccess(c42, i)),
				                  relTol,
				                  absTol));
			}

		}
		catch (CoordsException & e)
		{
			// cerr << "infinite component in /" << endl;
			assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de ostream & operator <<(ostream & out, const Coords<T, N> & c) ... ";
	{


		for (size_t t = 0; t < nbNumericTests; t++)
		{
			ostringstream out2;
			Coords<double, 2> c2(random(), random());

			out2 << c2;

			/*
			 * La chaîne attendue pour une coordonnée est :
			 * (x1, x2,    , xn)
			 */
			ostringstream oss2;
			oss2 << '(';

			size_t nbCoords = Coords<double, 2>::size();
			for (size_t i = 0; i < nbCoords; i++)
			{
				oss2 << dataAccess(c2, i);
				if (i < (nbCoords - 1))
				{
					oss2 << ", ";
				}
			}

			oss2 << ')';

			string expected = oss2.str();

			/*
			 * Le flux de sortie contient la chaine attendue
			 */
			assert(out2.str() == expected);

			expected.clear();

			Coords<double, 3> c3(random(), random(), random());
			ostringstream out3;

			out3 << c3;

			/*
			 * La chaîne attendue pour une coordonnée est :
			 * (x1, x2,    , xn)
			 */
			ostringstream oss3;
			oss3 << '(';

			nbCoords = Coords<double, 3>::size();
			for (size_t i = 0; i < nbCoords; i++)
			{
				oss3 << dataAccess(c3, i);
				if (i < (nbCoords - 1))
				{
					oss3 << ", ";
				}
			}

			oss3 << ')';

			expected = oss3.str();

			/*
			 * Le flux de sortie contient la chaine attendue
			 */
			assert(out3.str() == expected);

			expected.clear();

			Coords<double, 4> c4(random(), random(), random(), random());
			ostringstream out4;

			out4 << c4;

			/*
			 * La chaîne attendue pour une coordonnée est :
			 * (x1, x2,    , xn)
			 */
			ostringstream oss4;
			oss4 << '(';

			nbCoords = Coords<double, 4>::size();
			for (size_t i = 0; i < nbCoords; i++)
			{
				oss4 << dataAccess(c4, i);
				if (i < (nbCoords - 1))
				{
					oss4 << ", ";
				}
			}

			oss4 << ')';

			expected = oss4.str();

			/*
			 * Le flux de sortie contient la chaine attendue
			 */
			assert(out4.str() == expected);
		}
	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T dotProduct(const Coords<T, N> & c) const ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			Coords<double, 2> c22(random(), random());
			Coords<double, 3> c32(random(), random(), random());
			Coords<double, 4> c42(random(), random(), random(), random());

			double d2 = c21.dotProduct(c22);
			double d3 = c31.dotProduct(c32);
			double d4 = c41.dotProduct(c42);

			double sum2 = static_cast<double>(0);
			for (size_t i = 0; i < c21.size(); i++)
			{
				sum2 += dataAccess(c21, i) * dataAccess(c22, i);
			}

			assert(static_cast<double>(d2) == static_cast<double>(sum2));

			double sum3 = static_cast<double>(0);
			for (size_t i = 0; i < c31.size(); i++)
			{
				sum3 += dataAccess(c31, i) * dataAccess(c32, i);
			}

			assert(static_cast<double>(d3) == static_cast<double>(sum3));

			double sum4 = static_cast<double>(0);
			for (size_t i = 0; i < c41.size(); i++)
			{
				sum4 += dataAccess(c41, i) * dataAccess(c42, i);
			}

			assert(static_cast<double>(d4) == static_cast<double>(sum4));
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de T norm2() const ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c21(random(), random());
			Coords<double, 3> c31(random(), random(), random());
			Coords<double, 4> c41(random(), random(), random(), random());

			double d2 = c21.norm2();
			double d3 = c31.norm2();
			double d4 = c41.norm2();

			double sum2 = static_cast<double>(0);
			for (size_t i = 0; i < c21.size(); i++)
			{
				double val = dataAccess(c21, i);
				sum2 += val * val;
			}

			assert(static_cast<double>(d2) == static_cast<double>(sum2));

			double sum3 = static_cast<double>(0);
			for (size_t i = 0; i < c31.size(); i++)
			{
				double val = dataAccess(c31, i);
				sum3 += val * val;
			}

			assert(static_cast<double>(d3) == static_cast<double>(sum3));

			double sum4 = static_cast<double>(0);
			for (size_t i = 0; i < c41.size(); i++)
			{
				double val = dataAccess(c41, i);
				sum4 += val * val;
			}

			assert(static_cast<double>(d4) == static_cast<double>(sum4));
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & translate(const Coords<T, N> & c) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c2(random(), random());
			Coords<double, 3> c3(random(), random(), random());
			Coords<double, 4> c4(random(), random(), random(), random());

			Coords<double, 2> c2Saved(c2);
			Coords<double, 3> c3Saved(c3);
			Coords<double, 4> c4Saved(c4);

			Coords<double, 2> t21(random(), random());
			Coords<double, 3> t31(random(), random(), random());
			Coords<double, 4> t41(random(), random(), random(), random());
			Coords<double, 2> t22(random(), random());
			Coords<double, 3> t32(random(), random(), random());
			Coords<double, 4> t42(random(), random(), random(), random());

			c2.translate(t21).translate(t22);
			c3.translate(t31).translate(t32);
			c4.translate(t41).translate(t42);

			for (size_t i = 0; i < c2.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c2, i)),
				                  static_cast<double>(dataAccess(c2Saved, i) +
				                                      dataAccess(t21, i) +
				                                      dataAccess(t22, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c3.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c3, i)),
				                  static_cast<double>(dataAccess(c3Saved, i) +
				                                      dataAccess(t31, i) +
				                                      dataAccess(t32, i)),
				                  relTol,
				                  absTol));
			}

			for (size_t i = 0; i < c4.size(); i++)
			{
				assert(checkClose(static_cast<double>(dataAccess(c4, i)),
				                  static_cast<double>(dataAccess(c4Saved, i) +
				                                      dataAccess(t41, i) +
				                                      dataAccess(t42, i)),
				                  relTol,
				                  absTol));
			}
		}

	}
	cout << "Done" << endl;

	// ------------------------------------------------------------------------
	cout << "Test de Coords<T, N> & scale(const T & scale) ... ";
	{
		for (size_t t = 0; t < nbNumericTests; t++)
		{
			Coords<double, 2> c2(random(), random());
			Coords<double, 3> c3(random(), random(), random());
			Coords<double, 4> c4(random(), random(), random(), random());

			Coords<double, 2> c2Saved(c2);
			Coords<double, 3> c3Saved(c3);
			Coords<double, 4> c4Saved(c4);

			double s1(random());
			double s2(random());

			try
			{
				c2.scale(s1).scale(s2);

				for (size_t i = 0; i < c2.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c2, i)),
					                  static_cast<double>(dataAccess(c2Saved, i) *
					                                      s1 * s2),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c3.scale(s1).scale(s2);

				for (size_t i = 0; i < c3.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c3, i)),
					                  static_cast<double>(dataAccess(c3Saved, i) *
					                                      s1 * s2),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c4.scale(s1).scale(s2);
				for (size_t i = 0; i < c4.size(); i++)
				{
					assert(checkClose(static_cast<double>(dataAccess(c4, i)),
					                  static_cast<double>(dataAccess(c4Saved, i) *
					                                      s1 * s2),
					                  relTol,
					                  absTol));
				}
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			/*
			 * Checking infinite values
			 */
			double sinf(numeric_limits<double>::max());
			try
			{
				c2.scale(sinf);

				cerr << "Passed Infinite Scale";
				return EXIT_FAILURE;
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c3.scale(sinf);

				cerr << "Passed Infinite Scale";
				return EXIT_FAILURE;
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}

			try
			{
				c4.scale(sinf);

				cerr << "Passed Infinite Scale";
				return EXIT_FAILURE;
			}
			catch (CoordsException & e)
			{
				assert(e.getCause() == CoordsException::ExceptionCause::INFINITY_VALUE);
			}
		}
	}
	cout << "Done" << endl;

	return EXIT_SUCCESS;
}
